/* cygwin/fs.h
   Written by Chris January <chris@atomice.net>

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

/* These are Cygwin specific, however they should have the same values
 * as their Linux counterparts.
 */

#ifndef _CYGWIN_FS_H_
#define _CYGWIN_FS_H_

#define BLKRRPART    0x0000125f
#define BLKGETSIZE   0x00001260
#define BLKSSZGET    0x00001268
#define BLKIOMIN     0x00001278
#define BLKIOOPT     0x00001279
#define BLKALIGNOFF  0x0000127a
#define BLKPBSZGET   0x0000127b
#define BLKGETSIZE64 0x00041268

#endif

