/**
 * From https://github.com/raltnoeder/cppdsaext
 * Included under the GPLv2 license. See source project for other licensing options.
 *
 * Data structures and algorithms extensions
 *
 * @version 2018-05-16_001
 * @author  Robert Altnoeder (r.altnoeder@gmx.net)
 *
 * Copyright (C) 2012 - 2018 Robert ALTNOEDER
 */
#include <dsaext.h>

dsaext::DuplicateInsertException::~DuplicateInsertException() noexcept
{
}

dsaext::NumberFormatException::~NumberFormatException() noexcept
{
}
