#ifndef DEFAULT_TYPES_H
#define DEFAULT_TYPES_H

#include <cstddef>
#include <cstdint>

using std::nullptr_t;
using std::size_t;
using std::uint8_t;
using std::uint16_t;
using std::uint32_t;
using std::uint64_t;
using std::int8_t;
using std::int16_t;
using std::int32_t;
using std::int64_t;

#endif // DEFAULT_TYPES_H
