#ifndef GLOBALCOMMANDCONSTS_H
#define GLOBALCOMMANDCONSTS_H

#include <default_types.h>
#include <string>

namespace cmd_names
{
    extern const std::string    KEY_CMD_EXIT;
    extern const std::string    KEY_CMD_DISPLAY;
    extern const std::string    KEY_CMD_PAGE;
    extern const std::string    KEY_CMD_REFRESH;
    extern const std::string    KEY_CMD_COLORS;
    extern const std::string    KEY_CMD_CHARSET;
    extern const std::string    KEY_CMD_SELECT;
    extern const std::string    KEY_CMD_SELECT_ALL;
    extern const std::string    KEY_CMD_DESELECT;
    extern const std::string    KEY_CMD_DESELECT_ALL;
    extern const std::string    KEY_CMD_CLEAR_SELECTION;
    extern const std::string    KEY_CMD_CURSOR;
    extern const std::string    KEY_CMD_RESOURCE;
    extern const std::string    KEY_CMD_CONNECTION;
    extern const std::string    KEY_CMD_VOLUME;
    extern const std::string    KEY_CMD_MINOR_NR;
    extern const std::string    KEY_CMD_CLOSE;
}

#endif /* GLOBALCOMMANDCONSTS_H */
