#ifndef KEY_CODES_H
#define KEY_CODES_H

#include <default_types.h>

class KeyCodes
{
  public:
    static const uint32_t   ARROW_UP;
    static const uint32_t   ARROW_DOWN;
    static const uint32_t   ARROW_LEFT;
    static const uint32_t   ARROW_RIGHT;
    static const uint32_t   HOME;
    static const uint32_t   END;
    static const uint32_t   INSERT;
    static const uint32_t   DELETE;
    static const uint32_t   PG_UP;
    static const uint32_t   PG_DOWN;
    static const uint32_t   BACKSPACE;
    static const uint32_t   TAB;
    static const uint32_t   ENTER;
    static const uint32_t   FUNC_01;
    static const uint32_t   FUNC_02;
    static const uint32_t   FUNC_03;
    static const uint32_t   FUNC_04;
    static const uint32_t   FUNC_05;
    static const uint32_t   FUNC_06;
    static const uint32_t   FUNC_07;
    static const uint32_t   FUNC_08;
    static const uint32_t   FUNC_09;
    static const uint32_t   FUNC_10;
    static const uint32_t   FUNC_11;
    static const uint32_t   FUNC_12;
    static const uint32_t   MOUSE_EVENT;
    static const uint32_t   NONE;
};

#endif /* KEY_CODES_H */
