#include <terminal/GlobalCommandConsts.h>

namespace cmd_names
{
    // Command names must be all upper-case
    const std::string   KEY_CMD_EXIT("EXIT");
    const std::string   KEY_CMD_REFRESH("REFRESH");
    const std::string   KEY_CMD_PAGE("PAGE");
    const std::string   KEY_CMD_DISPLAY("DISPLAY");
    const std::string   KEY_CMD_COLORS("COLORS");
    const std::string   KEY_CMD_CHARSET("CHARSET");
    const std::string   KEY_CMD_SELECT("SELECT");
    const std::string   KEY_CMD_SELECT_ALL("SELECT-ALL");
    const std::string   KEY_CMD_DESELECT("DESELECT");
    const std::string   KEY_CMD_DESELECT_ALL("DESELECT-ALL");
    const std::string   KEY_CMD_CLEAR_SELECTION("CLEAR-SELECTION");
    const std::string   KEY_CMD_CURSOR("CURSOR");
    const std::string   KEY_CMD_RESOURCE("RESOURCE");
    const std::string   KEY_CMD_CONNECTION("CONNECTION");
    const std::string   KEY_CMD_VOLUME("VOLUME");
    const std::string   KEY_CMD_MINOR_NR("MINOR-NR");
    const std::string   KEY_CMD_CLOSE("CLOSE");
}
