#ifndef INPUTCHARCODES_H
#define INPUTCHARCODES_H

#include <default_types.h>

class InputCharCodes
{
  public:
    static const char   ENTER_1;
    static const char   ENTER_2;
    static const char   BACKSPACE;
    static const char   TAB;
};

#endif /* INPUTCHARCODES_H */
