#include <terminal/InputCharCodes.h>

const char InputCharCodes::ENTER_1      = 0xD;
const char InputCharCodes::ENTER_2      = 0xA;
const char InputCharCodes::TAB          = 0x9;
const char InputCharCodes::BACKSPACE    = 0x7F;
